/*
 * runmcConsoleProcess.java - A running process
 * :tabSize=8:indentSize=8:noTabs=false:
 * :folding=explicit:collapseFolds=1:
 *
 * Copyright (C) 1999, 2004 Slava Pestov
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package runmcconsole;

//{{{ Imports
import gnu.regexp.*;
import java.awt.Color;
import java.io.*;
import java.util.Stack;
import org.gjt.sp.jedit.*;
import org.gjt.sp.jedit.search.RESearchMatcher;
import org.gjt.sp.util.Log;
import errorlist.*;
//}}}

class runmcConsoleProcess
{
	//{{{ runmcConsoleProcess constructor
	runmcConsoleProcess(runmcConsole runmcconsole, Output output,
		Output error, String[] args, String[] env,
		SystemShell.runmcConsoleState runmcconsoleState, boolean foreground)
	{
		this.args = args;
		this.env = env;
		this.currentDirectory = runmcconsoleState.currentDirectory;

		if(foreground)
		{
			this.runmcconsole = runmcconsole;
			this.output = output;
			this.error = error;
			this.runmcconsoleState = runmcconsoleState;
		}

		try
		{
			pipeIn = new PipedInputStream();
			pipeOut = new PipedOutputStream(pipeIn);

			process = ProcessRunner.getProcessRunner()
				.exec(args,env,currentDirectory);
			stdin = new InputThread(this,
				process.getOutputStream());
			stdin.start();
			stdout = new StreamThread(this,
				process.getInputStream(),null);
			stdout.start();
			stderr = new StreamThread(this,
				process.getErrorStream(),
				runmcconsole.getErrorColor());
			stderr.start();
		}
		catch(Exception e)
		{
			String[] pp = { e.toString() };
			error.print(runmcconsole.getErrorColor(),
				jEdit.getProperty("runmcconsole.shell.error",pp));
			output.commandDone();
			error.commandDone();
			stop();
		}
	} //}}}

	//{{{ detach() method
	void detach()
	{
		if(runmcconsole != null)
		{
			Object[] pp = { args[0] };
			error.print(runmcconsole.getErrorColor(),
				jEdit.getProperty("runmcconsole.shell.detached",pp));
			output.commandDone();
			error.commandDone();
		}

		runmcconsoleState.process = null;
		runmcconsoleState = null;
		runmcconsole = null;
	} //}}}

	//{{{ stop() method
	void stop()
	{
		if(process != null)
		{
			stopped = true;

			stdin.abort();
			stdout.abort();
			stderr.abort();

			try
			{
				pipeOut.close();
			}
			catch(IOException e)
			{
				throw new RuntimeException(e);
			}

			process.destroy();
			process = null;

			if(runmcconsole != null)
			{
				Object[] pp = { args[0] };
				error.print(runmcconsole.getErrorColor(),
					jEdit.getProperty("runmcconsole.shell.killed",pp));
			}

			runmcConsolePlugin.finishErrorParsing(runmcconsole.getErrorSource());

			output.commandDone();
			error.commandDone();
		}

		if(runmcconsoleState != null)
			runmcconsoleState.process = null;
	} //}}}

	//{{{ isRunning() method
	boolean isRunning()
	{
		return (process != null);
	} //}}}
	
	//{{{ getExitStatus() method
	boolean getExitStatus()
	{
		return (exitCode == 0);
	} //}}}

	//{{{ getrunmcConsole() method
	runmcConsole getrunmcConsole()
	{
		return runmcconsole;
	} //}}}
	
	//{{{ getOutput() method
	Output getOutput()
	{
		return output;
	} //}}}
	
	//{{{ getErrorOutput() method
	Output getErrorOutput()
	{
		return error;
	} //}}}
	
	//{{{ getCurrentDirectory() method
	String getCurrentDirectory()
	{
		return currentDirectory;
	} //}}}
	
	//{{{ getPipeInput() method
	PipedInputStream getPipeInput()
	{
		return pipeIn;
	} //}}}
	
	//{{{ getPipeOutput() method
	PipedOutputStream getPipeOutput()
	{
		return pipeOut;
	} //}}}
	
	//{{{ threadDone() method
	synchronized void threadDone()
	{
		threadDoneCount++;
		if(process == null)
			return;

		if(!stopped)
		{
			// we don't want unkillable processes to hang
			// jEdit

			try
			{
				exitCode = process.waitFor();
			}
			catch(InterruptedException e)
			{
				Log.log(Log.ERROR,this,e);
				notifyAll();
				return;
			}
			
			try
			{
				pipeOut.close();
			}
			catch(IOException io)
			{
			}
		}

		if(threadDoneCount == 3)
		{
			if(runmcconsole != null && output != null && error != null)
			{
				Object[] pp = { args[0], new Integer(exitCode) };

				String msg = jEdit.getProperty("runmcconsole.shell.exited",pp);

				if(exitCode == 0)
					error.print(runmcconsole.getInfoColor(),msg);
				else
					error.print(runmcconsole.getErrorColor(),msg);

				runmcConsolePlugin.finishErrorParsing(
					runmcconsole.getErrorSource());

				output.commandDone();
				error.commandDone();
				
				jEdit.checkBufferStatus(jEdit.getActiveView());
			}

			process = null;

			if(runmcconsoleState != null)
				runmcconsoleState.process = null;

			notifyAll();
		}
	} //}}}
	
	//{{{ Private members
	private SystemShell.runmcConsoleState runmcconsoleState;
	private String currentDirectory;
	private runmcConsole runmcconsole;
	private Output output;
	private Output error;
	private String[] args;
	private String[] env;
	private Process process;
	private InputThread stdin;
	private StreamThread stdout;
	private StreamThread stderr;
	private int threadDoneCount;
	private int exitCode;
	private boolean stopped;

	/* AWT thread writes stdin to this pipe, and the input thread
	writes it to the process. */
	private PipedInputStream pipeIn;
	private PipedOutputStream pipeOut;
	//}}}
}
