/*
 *  ButtonBar in org.jpws.front.util
 *  file: ButtonBar.java
 * 
 *  Project Jpws-Front
 *  @author Wolfgang Keller
 *  Created 28.09.2004
 *  Version
 * 
 *  Copyright (c) 2005 by Wolfgang Keller, Munich, Germany
 * 
 This program is not freeware software but copyright protected to the author(s)
 stated above. However, you can use, redistribute and/or modify it under the terms 
 of the GNU General Public License as published by the Free Software Foundation, 
 version 2 of the License.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with
 this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 Place - Suite 330, Boston, MA 02111-1307, USA, or go to
 http://www.gnu.org/copyleft/gpl.html.
 */
package JRunMC;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;


/**
 * Creates a container suitable for holding buttons.  The buttons are added 
 * using the {@link #add(JButton)}
 * method and are arranged as a horizontal strip with each button the same size 
 * as the widest button.  If
 * the button bar is put into a container wider than its natural size, the 
 * buttons are not resized, but
 * horizontal spacing is added to centre the buttons within the area.
 * 
 * @author Kevin Preece
 */
public class ButtonBar extends JPanel
{
	public static final int		LEFT		= 0;
	public static final int		TOP			= 0;
	public static final int		RIGHT		= 1;
	public static final int		BOTTOM		= 1;
	public static final int		CENTRE		= 2;
	public static final int		CENTER		= 2;
	public static final int		HORIZONTAL	= 3;
	public static final int		VERTICAL	= 4;

	private JPanel		buttonPanel;

	/**
	 * Creates a button bar that is initially empty with the default orientation of
	 * <code>HORIZONTAL</code> and alignment of <code>CENTRE</code>.
	 */
	public ButtonBar()
	{
		this( HORIZONTAL, CENTRE );
	}

	/**
	 * Creates a bar that JButton buttons are added to.  All buttons will be resized
	 * so that they are equal.
	 * 
	 * @param orientation  the orientation of the button bar <code>HORIZONTAL</code> or <code>VERTICAL</code>.
	 * @param alignment    the button bar alignment - <code>LEFT</code>, <code>CENTRE</code> or <code>RIGHT</code>.
    * @param margin <code>Insets</code> to define the empty border area of the bar 
	 */
	public ButtonBar( int orientation, int alignment, Insets margin )
	{
		super();

		JPanel				layout;
		GridBagConstraints	c;
		GridLayout			buttonLayout;

		if ( orientation == HORIZONTAL )
		{
			// rows, cols, hgap, vgap
			buttonLayout = new GridLayout( 1, 0, 8, 0 );
		}
		else if ( orientation == VERTICAL )
		{
			// rows, cols, hgap, vgap
			buttonLayout = new GridLayout( 0, 1, 0, 4 );
		}
		else
		{
			throw new IllegalArgumentException();
		}

		if ( (alignment != LEFT) && (alignment != CENTRE) && (alignment != RIGHT) )
		{
			throw new IllegalArgumentException();
		}

		layout			= new JPanel( new GridBagLayout() );
		buttonPanel		= new JPanel( buttonLayout );
		c				= new GridBagConstraints();

		if ( (alignment == CENTRE) || (alignment == RIGHT) )
		{
			c.fill			= GridBagConstraints.HORIZONTAL;
			layout.add( spacerPanel( 10, 10 ), c );
		}
		
		c.fill			= GridBagConstraints.NONE;
		layout.add( buttonPanel, c );

		if ( (alignment == CENTRE) || (alignment == LEFT) )
		{
			c.fill			= GridBagConstraints.HORIZONTAL;
			layout.add( spacerPanel( 10, 10 ), c );
		}

		this.setBorder( new EmptyBorder( margin ) );
		this.add( layout );
	}

   private JPanel spacerPanel ( int width, int height )
   {
      JPanel panel;
      
      panel = new JPanel();
      panel.setPreferredSize( new Dimension( width, height ) );
      return panel;
   }
   
   /**
    * Creates a bar that JButton buttons are added to.  All buttons will be resized
    * so that they are equal. There is a default margin for the border area.
    * 
    * @param orientation  the orientation of the button bar <code>HORIZONTAL</code>
    *        or <code>VERTICAL</code>.
    * @param alignment    the button bar alignment - <code>LEFT</code>, 
    *        <code>CENTRE</code> or <code>RIGHT</code>. 
    */
   public ButtonBar( int orientation, int alignment )
   {
      this( orientation, alignment, new Insets( 12, 0, 6, 0 ) );
   }

	/**
	 * Adds a button to this button bar.
	 * 
	 * @param button the button to add.
	 */
	public void add( JButton button )
	{
		buttonPanel.add( button );
	}
}
